from typing import Any, Dict, List, Optional, Tuple

from transformers import pipeline


class TextLabellingTransformation:
    """
    Text labelling transformation

    Input Format: str
    Output Format: {'labels': [<str>], scores: [<float>]}
    """

    def __init__(self, model: str, input_field: str, output_field: str, labels: List[str],
                 gpu_device: Optional[int] = None):
        self._model = model
        self._input_field = input_field
        self._output_field = output_field
        self._labels = labels
        self._gpu_device = gpu_device
        self._pipeline = pipeline(model=self._model, device=self._gpu_device,
                                  top_k=len(labels),
                                  task='zero-shot-classification')

    def transform(self, data_list: List[Dict[str, Any]]):
        indexed_inputs = self._get_indexed_inputs(data_list)
        model_inputs = [d for _, d in indexed_inputs]
        if not model_inputs:
            return
        model_outputs = self._pipeline(model_inputs, candidate_labels=self._labels)
        self._set_all_scores(data_list, indexed_inputs, model_outputs)

    def _get_indexed_inputs(self, data_list: List[Dict[str, Any]]) -> List[Tuple[int, str]]:
        indexed_inputs = []
        for idx, data in enumerate(data_list):
            if self._input_field in data and data[self._input_field] is not None:
                indexed_inputs.append((idx, data[self._input_field],))

        return indexed_inputs

    def _set_all_scores(self, data_list: List[Dict[str, Any]], indexed_inputs: List[Tuple[int, str]],
                        model_outputs: List[Dict]):
        for idx, (data_index, _) in enumerate(indexed_inputs):
            model_output = model_outputs[idx]
            model_output.pop('sequence')
            data_list[data_index][self._output_field] = model_output



from typing import Any, Dict, List, Optional, Tuple

from transformers import pipeline

from src.gdaicommons.utils import normalize_model_output


class FillMaskTransformation:
    """
    Fill Mask Transformation

    Input Format: <str>
        Examples: "Sample sentence <mask_placeholder>", "Sample sentence [MASK]"
    Output Format: [{'score': <float>, 'token': <int>, 'token_str': <str>}, ...]
    """

    def __init__(self, model: str, input_field: str, output_field: str,
                 mask_placeholder: str = '[MASK]',
                 mask_targets: Optional[List[str]] = None,
                 top_k_scores: Optional[int] = 5,
                 gpu_device: Optional[int] = None):
        self._model = model
        self._input_field = input_field
        self._output_field = output_field
        self._mask_placeholder = mask_placeholder
        self._mask_targets = mask_targets
        self._top_k_scores = top_k_scores if top_k_scores else 5
        self._gpu_device = gpu_device
        self._pipeline = pipeline(model=self._model, device=self._gpu_device,
                                  top_k=self._top_k_scores,
                                  task='fill-mask')

    def transform(self, data_list: List[Dict[str, Any]]):
        indexed_inputs = self._get_indexed_inputs(data_list)
        model_inputs = [d for _, d in indexed_inputs]
        if not model_inputs:
            return
        if self._mask_placeholder != '[MASK]':
            model_inputs = [str(d).replace(self._mask_placeholder, '[MASK]') for d in model_inputs]
        if self._mask_targets:
            model_outputs = self._pipeline(model_inputs, targets=self._mask_targets)
        else:
            model_outputs = self._pipeline(model_inputs)
        model_outputs = normalize_model_output(model_outputs)
        self._set_all_scores(data_list, indexed_inputs, model_outputs)

    def _get_indexed_inputs(self, data_list: List[Dict[str, Any]]) -> List[Tuple[int, str]]:
        indexed_inputs = []
        for idx, data in enumerate(data_list):
            if self._input_field in data and data[self._input_field] is not None:
                indexed_inputs.append((idx, data[self._input_field],))

        return indexed_inputs

    def _set_all_scores(self, data_list: List[Dict[str, Any]], indexed_inputs: List[Tuple[int, str]],
                        model_outputs: List[List[Dict]]):
        for idx, (data_index, _) in enumerate(indexed_inputs):
            model_output = model_outputs[idx]
            model_output = [FillMaskTransformation._remove_sequence(d) for d in model_output]
            data_list[data_index][self._output_field] = model_output

    @staticmethod
    def _remove_sequence(output: dict):
        if 'sequence' in output:
            output.pop('sequence')
        return output



import unittest

from tests.fill_mask import FillMaskTest
from tests.text2text_generation import Text2TextGenerationTest
from tests.text_classification import TextClassificationTest
from tests.text_generation import TextGenerationTest
from tests.text_labelling import TextLabellingTest
from tests.text_summarization import TextSummarizationTest

test_suite = unittest.TestSuite()
test_suite.addTest(unittest.makeSuite(FillMaskTest))
test_suite.addTest(unittest.makeSuite(Text2TextGenerationTest))
test_suite.addTest(unittest.makeSuite(TextClassificationTest))
test_suite.addTest(unittest.makeSuite(TextGenerationTest))
test_suite.addTest(unittest.makeSuite(TextLabellingTest))
test_suite.addTest(unittest.makeSuite(TextSummarizationTest))

if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(test_suite)



from typing import Dict, List, Optional, Union


def normalize_model_output(output: Union[List[List[Dict]], List[Dict]]) -> List[List[Dict]]:
    if not output:
        return []

    if isinstance(output[0], list):
        return output
    else:
        return [output]


def parse_parameter(param: Optional[str], cast_class):
    if not param:
        return param

    return cast_class(param)


def blank_if_null(param: str) -> str:
    if param is None:
        return ""
    return param



from gdtransform.transform import transformation_builder

from .fill_mask import FillMaskTransformation
from .text2text_generation import Text2TextGenerationTransformation
from .text_classification import TextClassificationTransformation
from .text_generation import TextGenerationTransformation
from .text_labelling import TextLabellingTransformation
from .text_summarization import TextSummarizationTransformation
from ..constants import VALUE_SEPARATOR
from ..utils import blank_if_null, parse_parameter


@transformation_builder(name='text-classification-builder', is_batch=True)
def text_classification(*args, **kwargs):
    model = kwargs['model']
    input_field = kwargs['input_field']
    output_field = kwargs['output_field']
    gpu_device = int(kwargs.get('gpu_device')) if 'gpu_device' in kwargs and kwargs['gpu_device'] else None

    text_classifier = TextClassificationTransformation(model, input_field, output_field,
                                                       gpu_device=gpu_device)

    return text_classifier.transform


@transformation_builder(name='text-labelling-builder', is_batch=True)
def text_labelling(*args, **kwargs):
    model = kwargs['model']
    input_field = kwargs['input_field']
    output_field = kwargs['output_field']
    labels = kwargs['labels']
    labels = str(labels).split(sep=VALUE_SEPARATOR)
    labels = [l.strip() for l in labels]
    gpu_device = int(kwargs.get('gpu_device')) if 'gpu_device' in kwargs and kwargs['gpu_device'] else None

    text_labeller = TextLabellingTransformation(model, input_field, output_field, labels,
                                                gpu_device=gpu_device)
    return text_labeller.transform


@transformation_builder(name='fill-mask-builder', is_batch=True)
def fill_mask(*args, **kwargs):
    model = kwargs['model']
    input_field = kwargs['input_field']
    output_field = kwargs['output_field']
    mask_placeholder = kwargs.get('mask_placeholder', '[MASK]')
    mask_targets = kwargs.get('mask_targets', None)
    if mask_targets:
        mask_targets = str(mask_targets).split(sep=VALUE_SEPARATOR)
        mask_targets = [t.strip() for t in mask_targets]
    gpu_device = int(kwargs.get('gpu_device')) if 'gpu_device' in kwargs and kwargs['gpu_device'] else None

    fill_mask_transformation = FillMaskTransformation(model, input_field, output_field,
                                                      mask_placeholder=mask_placeholder,
                                                      mask_targets=mask_targets,
                                                      top_k_scores=5,
                                                      gpu_device=gpu_device)
    return fill_mask_transformation.transform


@transformation_builder(name='text-generation-builder', is_batch=True)
def text_generation(*args, **kwargs):
    model = kwargs['model']
    input_field = kwargs['input_field']
    output_field = kwargs['output_field']
    max_length = parse_parameter(kwargs.get('max_length', None), int)
    max_new_tokens = parse_parameter(kwargs.get('max_new_tokens', None), int)
    min_length = parse_parameter(kwargs.get('min_length', None), int)
    min_new_tokens = parse_parameter(kwargs.get('min_new_tokens', None), int)
    do_sample = blank_if_null(kwargs.get('do_sample', 'false')).lower() in ['true']
    num_beams = parse_parameter(kwargs.get('num_beams', '1'), int)
    temperature = parse_parameter(kwargs.get('temperature', '1.0'), float)
    top_k = parse_parameter(kwargs.get('top_k', '50'), int)
    gpu_device = int(kwargs.get('gpu_device')) if 'gpu_device' in kwargs and kwargs['gpu_device'] else None

    text_generation_transformation = TextGenerationTransformation(model, input_field, output_field,
                                                                  max_length=max_length,
                                                                  max_new_tokens=max_new_tokens,
                                                                  min_length=min_length,
                                                                  min_new_tokens=min_new_tokens,
                                                                  do_sample=do_sample,
                                                                  num_beams=num_beams,
                                                                  temperature=temperature,
                                                                  top_k=top_k,
                                                                  gpu_device=gpu_device)
    return text_generation_transformation.transform


@transformation_builder(name='text-summarization-builder', is_batch=True)
def text_summarization(*args, **kwargs):
    model = kwargs['model']
    input_field = kwargs['input_field']
    output_field = kwargs['output_field']
    max_length = parse_parameter(kwargs.get('max_length', None), int)
    max_new_tokens = parse_parameter(kwargs.get('max_new_tokens', None), int)
    min_length = parse_parameter(kwargs.get('min_length', None), int)
    min_new_tokens = parse_parameter(kwargs.get('min_new_tokens', None), int)
    do_sample = blank_if_null(kwargs.get('do_sample', 'false')).lower() in ['true']
    num_beams = parse_parameter(kwargs.get('num_beams', '1'), int)
    temperature = parse_parameter(kwargs.get('temperature', '1.0'), float)
    top_k = parse_parameter(kwargs.get('top_k', '50'), int)
    gpu_device = int(kwargs.get('gpu_device')) if 'gpu_device' in kwargs and kwargs['gpu_device'] else None

    text_summarization_transformation = TextSummarizationTransformation(model, input_field, output_field,
                                                                        max_length=max_length,
                                                                        max_new_tokens=max_new_tokens,
                                                                        min_length=min_length,
                                                                        min_new_tokens=min_new_tokens,
                                                                        do_sample=do_sample,
                                                                        num_beams=num_beams,
                                                                        temperature=temperature,
                                                                        top_k=top_k,
                                                                        gpu_device=gpu_device)
    return text_summarization_transformation.transform


@transformation_builder(name='text2text-generation-builder', is_batch=True)
def text2text_generation(*args, **kwargs):
    model = kwargs['model']
    input_field = kwargs['input_field']
    output_field = kwargs['output_field']
    max_length = parse_parameter(kwargs.get('max_length', None), int)
    max_new_tokens = parse_parameter(kwargs.get('max_new_tokens', None), int)
    min_length = parse_parameter(kwargs.get('min_length', None), int)
    min_new_tokens = parse_parameter(kwargs.get('min_new_tokens', None), int)
    do_sample = blank_if_null(kwargs.get('do_sample', 'false')).lower() in ['true']
    num_beams = parse_parameter(kwargs.get('num_beams', '1'), int)
    temperature = parse_parameter(kwargs.get('temperature', '1.0'), float)
    top_k = parse_parameter(kwargs.get('top_k', '50'), int)
    gpu_device = int(kwargs.get('gpu_device')) if 'gpu_device' in kwargs and kwargs['gpu_device'] else None

    text2text_generation_transformation = Text2TextGenerationTransformation(model, input_field, output_field,
                                                                            max_length=max_length,
                                                                            max_new_tokens=max_new_tokens,
                                                                            min_length=min_length,
                                                                            min_new_tokens=min_new_tokens,
                                                                            do_sample=do_sample,
                                                                            num_beams=num_beams,
                                                                            temperature=temperature,
                                                                            top_k=top_k,
                                                                            gpu_device=gpu_device)
    return text2text_generation_transformation.transform



from typing import Any, Dict, List, Optional, Tuple

from transformers import pipeline, set_seed


class TextSummarizationTransformation:
    """
    Text Summarization Transformation

    Input Format: <str>
    Output Format: <str>
    """

    def __init__(self, model: str, input_field: str, output_field: str,
                 max_length: Optional[int] = None,
                 max_new_tokens: Optional[int] = None,
                 min_length: Optional[int] = None,
                 min_new_tokens: Optional[int] = None,
                 do_sample: Optional[bool] = False,
                 num_beams: Optional[int] = 1,
                 temperature: Optional[float] = 1.0,
                 top_k: Optional[int] = 50,
                 gpu_device: Optional[int] = None):
        self._model = model
        self._input_field = input_field
        self._output_field = output_field
        self._gpu_device = gpu_device
        model_kwargs = {}
        if max_length:
            model_kwargs['max_length'] = max_length,
        if max_new_tokens:
            model_kwargs['max_new_tokens'] = max_new_tokens
        if min_length:
            model_kwargs['min_length'] = min_length
        if min_new_tokens:
            model_kwargs['min_new_tokens'] = min_new_tokens
        model_kwargs['do_sample'] = True if do_sample else False
        model_kwargs['num_beams'] = num_beams if num_beams else 1
        model_kwargs['temperature'] = float(temperature) if temperature else 1.0
        model_kwargs['top_k'] = top_k if top_k else 50
        set_seed(0)  # setting globally; might conflict with other models
        self._pipeline = pipeline(model=self._model, device=self._gpu_device,
                                  task='summarization',
                                  **model_kwargs)

    def transform(self, data_list: List[Dict[str, Any]]):
        indexed_inputs = self._get_indexed_inputs(data_list)
        model_inputs = [d for _, d in indexed_inputs]
        if not model_inputs:
            return
        model_outputs = self._pipeline(model_inputs)
        self._set_all_scores(data_list, indexed_inputs, model_outputs)

    def _get_indexed_inputs(self, data_list: List[Dict[str, Any]]) -> List[Tuple[int, str]]:
        indexed_inputs = []
        for idx, data in enumerate(data_list):
            if self._input_field in data and data[self._input_field] is not None:
                indexed_inputs.append((idx, data[self._input_field],))

        return indexed_inputs

    def _set_all_scores(self, data_list: List[Dict[str, Any]], indexed_inputs: List[Tuple[int, str]],
                        model_outputs: List[Dict[str, str]]):
        for idx, (data_index, _) in enumerate(indexed_inputs):
            model_output = model_outputs[idx]
            data_list[data_index][self._output_field] = model_output.get('summary_text', '')
